@@ -27,7 +27,7 @@ pub struct HeightAndSize {
 
 #[derive(Clone, Debug)]
 pub(crate) struct Node<K: Ord + Clone, V: Clone, const SIZE: usize> {
-    elts: Arc<Chunk<K, V, SIZE>>,
+    elts: Chunk<K, V, SIZE>,
     min_key: K,
     max_key: K,
     left: Tree<K, V, SIZE>,
@@ -260,7 +260,7 @@ where
             let (visited, current) = self.stack[top];
             if visited {
                 if self.any_elts_in_bounds(current) {
-                    self.elts = Some((&(*current.elts)).into_iter());
+                    self.elts = Some((&current.elts).into_iter());
                 }
                 self.stack.pop();
                 match current.right {
@@ -323,7 +323,7 @@ where
             let (visited, current) = self.stack_rev[top];
             if visited {
                 if self.any_elts_in_bounds(current) {
-                    self.elts_rev = Some((&(*current.elts)).into_iter());
+                    self.elts_rev = Some((&current.elts).into_iter());
                 }
                 self.stack_rev.pop();
                 match current.left {
@@ -432,14 +432,14 @@ where
         }
     }
 
-    fn add_min_elts(&self, elts: &Arc<Chunk<K, V, SIZE>>) -> Self {
+    fn add_min_elts(&self, elts: &Chunk<K, V, SIZE>) -> Self {
         match self {
             Tree::Empty => Tree::create(&Tree::Empty, elts.clone(), &Tree::Empty),
             Tree::Node(ref n) => Tree::bal(&n.left.add_min_elts(elts), &n.elts, &n.right),
         }
     }
 
-    fn add_max_elts(&self, elts: &Arc<Chunk<K, V, SIZE>>) -> Self {
+    fn add_max_elts(&self, elts: &Chunk<K, V, SIZE>) -> Self {
         match self {
             Tree::Empty => Tree::create(&Tree::Empty, elts.clone(), &Tree::Empty),
             Tree::Node(ref n) => Tree::bal(&n.left, &n.elts, &n.right.add_max_elts(elts)),
@@ -451,7 +451,7 @@ where
     // a balanced tree.
     fn join(
         l: &Tree<K, V, SIZE>,
-        elts: &Arc<Chunk<K, V, SIZE>>,
+        elts: &Chunk<K, V, SIZE>,
         r: &Tree<K, V, SIZE>,
     ) -> Self {
         match (l, r) {
@@ -459,9 +459,9 @@ where
             (_, Tree::Empty) => l.add_max_elts(elts),
             (Tree::Node(ref ln), Tree::Node(ref rn)) => {
                 let (ln_height, rn_height) = (ln.height(), rn.height());
-                if ln_height > rn_height + 2 {
+                if ln_height > rn_height + 1 {
                     Tree::bal(&ln.left, &ln.elts, &Tree::join(&ln.right, elts, r))
-                } else if rn_height > ln_height + 2 {
+                } else if rn_height > ln_height + 1 {
                     Tree::bal(&Tree::join(l, elts, &rn.left), &rn.elts, &rn.right)
                 } else {
                     Tree::create(l, elts.clone(), r)
@@ -475,7 +475,7 @@ where
     /// if there is a possible intersection return the intersecting
     /// chunk. In the case of an intersection there may also be an
     /// intersection at the left and/or right nodes.
-    fn split(&self, vmin: &K, vmax: &K) -> (Self, Option<Arc<Chunk<K, V, SIZE>>>, Self) {
+    fn split(&self, vmin: &K, vmax: &K) -> (Self, Option<Chunk<K, V, SIZE>>, Self) {
         match self {
             Tree::Empty => (Tree::Empty, None, Tree::Empty),
             Tree::Node(ref n) => {
@@ -666,7 +666,7 @@ where
 
     fn create(
         l: &Tree<K, V, SIZE>,
-        elts: Arc<Chunk<K, V, SIZE>>,
+        elts: Chunk<K, V, SIZE>,
         r: &Tree<K, V, SIZE>,
     ) -> Self {
         let (min_key, max_key) = elts.min_max_key().unwrap();
@@ -690,11 +690,40 @@ where
         (hl <= hr + 1) && (hr <= hl + 1)
     }
 
-    fn bal(
-        l: &Tree<K, V, SIZE>,
-        elts: &Arc<Chunk<K, V, SIZE>>,
-        r: &Tree<K, V, SIZE>,
-    ) -> Self {
+    fn compact(self) -> Self {
+        match self {
+            Tree::Empty => self,
+            Tree::Node(ref tn) => {
+                let len = tn.elts.len();
+                if len > SIZE >> 1 {
+                    self
+                } else {
+                    match tn.right.min_elts() {
+                        None => self,
+                        Some(chunk) => {
+                            let n = SIZE - len;
+                            let to_add =
+                                chunk.into_iter().map(|(k, v)| (k.clone(), v.clone()));
+                            let overflow = chunk
+                                .into_iter()
+                                .skip(n)
+                                .map(|(k, v)| (k.clone(), v.clone()));
+                            let elts = tn.elts.append(to_add);
+                            let t =
+                                Tree::bal(&tn.left, &elts, &tn.right.remove_min_elts());
+                            if n >= chunk.len() {
+                                t
+                            } else {
+                                t.insert_many(overflow)
+                            }
+                        }
+                    }
+                }
+            }
+        }
+    }
+
+    fn bal(l: &Tree<K, V, SIZE>, elts: &Chunk<K, V, SIZE>, r: &Tree<K, V, SIZE>) -> Self {
         let (hl, hr) = (l.height(), r.height());
         if hl > hr + 1 {
             match *l {
@@ -706,6 +735,7 @@ where
                             ln.elts.clone(),
                             &Tree::create(&ln.right, elts.clone(), r),
                         )
+                        .compact()
                     } else {
                         match ln.right {
                             Tree::Empty => panic!("tree heights wrong"),
@@ -713,7 +743,8 @@ where
                                 &Tree::create(&ln.left, ln.elts.clone(), &lrn.left),
                                 lrn.elts.clone(),
                                 &Tree::create(&lrn.right, elts.clone(), r),
-                            ),
+                            )
+                            .compact(),
                         }
                     }
                 }
@@ -728,6 +759,7 @@ where
                             rn.elts.clone(),
                             &rn.right,
                         )
+                        .compact()
                     } else {
                         match rn.left {
                             Tree::Empty => panic!("tree heights are wrong"),
@@ -735,13 +767,14 @@ where
                                 &Tree::create(l, elts.clone(), &rln.left),
                                 rln.elts.clone(),
                                 &Tree::create(&rln.right, rn.elts.clone(), &rn.right),
-                            ),
+                            )
+                            .compact(),
                         }
                     }
                 }
             }
         } else {
-            Tree::create(l, elts.clone(), r)
+            Tree::create(l, elts.clone(), r).compact()
         }
     }
 
@@ -755,11 +788,9 @@ where
             return self.clone();
         }
         match self {
-            &Tree::Empty => Tree::create(
-                &Tree::Empty,
-                Arc::new(Chunk::create_with(chunk, f)),
-                &Tree::Empty,
-            ),
+            &Tree::Empty => {
+                Tree::create(&Tree::Empty, Chunk::create_with(chunk, f), &Tree::Empty)
+            }
             &Tree::Node(ref tn) => {
                 let leaf = match (&tn.left, &tn.right) {
                     (&Tree::Empty, &Tree::Empty) => true,
@@ -846,11 +877,8 @@ where
             Tree::Empty => match f(q, d, None) {
                 None => None,
                 Some((k, v)) => {
-                    *self = Tree::create(
-                        &Tree::Empty,
-                        Arc::new(Chunk::singleton(k, v)),
-                        &Tree::Empty,
-                    );
+                    *self =
+                        Tree::create(&Tree::Empty, Chunk::singleton(k, v), &Tree::Empty);
                     None
                 }
             },
@@ -860,11 +888,11 @@ where
                     (&Tree::Empty, &Tree::Empty) => true,
                     (_, _) => false,
                 };
-                match Arc::make_mut(&mut tn.elts).update_mut(q, d, leaf, f) {
+                match tn.elts.update_mut(q, d, leaf, f) {
                     MutUpdate::UpdateLeft(k, d) => {
                         let prev = tn.left.update_cow(k, d, f);
                         if !Tree::in_bal(&tn.left, &tn.right) {
-                            *self = Tree::bal(&tn.left, &tn.elts, &tn.right);
+                            *self = Tree::bal(&tn.left, &tn.elts, &tn.right)
                         } else {
                             tn.mutated();
                         }
@@ -873,7 +901,7 @@ where
                     MutUpdate::UpdateRight(k, d) => {
                         let prev = tn.right.update_cow(k, d, f);
                         if !Tree::in_bal(&tn.left, &tn.right) {
-                            *self = Tree::bal(&tn.left, &tn.elts, &tn.right);
+                            *self = Tree::bal(&tn.left, &tn.elts, &tn.right)
                         } else {
                             tn.mutated();
                         }
@@ -921,11 +949,7 @@ where
             Tree::Empty => match f(q, d, None) {
                 None => (self.clone(), None),
                 Some((k, v)) => (
-                    Tree::create(
-                        &Tree::Empty,
-                        Arc::new(Chunk::singleton(k, v)),
-                        &Tree::Empty,
-                    ),
+                    Tree::create(&Tree::Empty, Chunk::singleton(k, v), &Tree::Empty),
                     None,
                 ),
             },
@@ -952,10 +976,7 @@ where
                             if elts.len() == 0 {
                                 (Tree::concat(&tn.left, &tn.right), previous)
                             } else {
-                                (
-                                    Tree::create(&tn.left, Arc::new(elts), &tn.right),
-                                    previous,
-                                )
+                                (Tree::create(&tn.left, elts, &tn.right), previous)
                             }
                         }
                         Some((ovk, ovv)) => {
@@ -980,10 +1001,10 @@ where
         self.update_cow(k, v, &mut |k, v, _| Some((k, v)))
     }
 
-    fn min_elts<'a>(&'a self) -> Option<&'a Arc<Chunk<K, V, SIZE>>> {
+    fn min_elts<'a>(&'a self) -> Option<&'a Chunk<K, V, SIZE>> {
         match self {
-            &Tree::Empty => None,
-            &Tree::Node(ref tn) => match tn.left {
+            Tree::Empty => None,
+            Tree::Node(ref tn) => match tn.left {
                 Tree::Empty => Some(&tn.elts),
                 Tree::Node(_) => tn.left.min_elts(),
             },
@@ -992,8 +1013,8 @@ where
 
     fn remove_min_elts(&self) -> Self {
         match self {
-            &Tree::Empty => panic!("remove min elt"),
-            &Tree::Node(ref tn) => match tn.left {
+            Tree::Empty => panic!("remove min elt"),
+            Tree::Node(ref tn) => match tn.left {
                 Tree::Empty => tn.right.clone(),
                 Tree::Node(_) => {
                     Tree::bal(&tn.left.remove_min_elts(), &tn.elts, &tn.right)
@@ -1004,8 +1025,8 @@ where
 
     fn concat(l: &Tree<K, V, SIZE>, r: &Tree<K, V, SIZE>) -> Tree<K, V, SIZE> {
         match (l, r) {
-            (&Tree::Empty, _) => r.clone(),
-            (_, &Tree::Empty) => l.clone(),
+            (Tree::Empty, _) => r.clone(),
+            (_, Tree::Empty) => l.clone(),
             (_, _) => {
                 let elts = r.min_elts().unwrap();
                 Tree::bal(l, elts, &r.remove_min_elts())
@@ -1027,7 +1048,7 @@ where
                     if elts.len() == 0 {
                         (Tree::concat(&tn.left, &tn.right), Some(p))
                     } else {
-                        (Tree::create(&tn.left, Arc::new(elts), &tn.right), Some(p))
+                        (Tree::create(&tn.left, elts, &tn.right), Some(p))
                     }
                 }
                 Loc::InLeft => {
@@ -1053,7 +1074,7 @@ where
                 match tn.elts.get(k) {
                     Loc::NotPresent(_) => None,
                     Loc::Here(i) => {
-                        let (_, p) = Arc::make_mut(&mut tn.elts).remove_elt_at_mut(i);
+                        let (_, p) = tn.elts.remove_elt_at_mut(i);
                         if tn.elts.len() == 0 {
                             *self = Tree::concat(&tn.left, &tn.right);
                             Some(p)
@@ -1074,7 +1095,7 @@ where
                     Loc::InRight => {
                         let p = tn.right.remove_cow(k);
                         if !Tree::in_bal(&tn.left, &tn.right) {
-                            *self = Tree::bal(&tn.left, &tn.elts, &tn.right)
+                            *self = Tree::bal(&tn.left, &tn.elts, &tn.right);
                         } else {
                             tn.mutated()
                         }
